#!/usr/bin/ruby -w

#TH 2009-11-15,17 
############################################################################
#    Copyright (C) 2009 by Thomas Hockne                                   #
#                                                                          #
#                                                                          #
#    This program is free software; you can redistribute it and/or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################
#2009-12-12 ver 0.2

require 'open-uri'
include Math

unless ARGV[0]
	puts "Usage: #{$0} report-time ( 0=this hour 1= last hour etc ) [lat long radius ] "
	puts "With just one argument, report-time, you get a gpxfile with world-wide ship reports"
	puts "for the given time span 0 = last hour, 1 = last two hours etc. If you additionally"
	puts "give a lat,long and radius, the output will be restricted to ships within the specified circle"
	puts "Report format for each ship: Wind direction, wind speed, time, air preassure."
	exit(false)
 end
 
ships = 0
 
 if ARGV[0] != nil
   time_span = ARGV[0].to_i
 end 
 
  if ARGV[1] != nil
   latC = ARGV[1].to_f
 end
 if ARGV[2] != nil
   longC = ARGV[2].to_f
 end
  if ARGV[3] != nil
   radius = ARGV[3].to_f
 end
 
 #ruby-fied c++ code from OpenCPN
def dist (slat1,slon1,slat2,slon2)

 lon1 = slon1.to_f * PI/180
 lon2 = slon2.to_f * PI/180
 lat1 = slat1.to_f * PI/180
 lat2 = slat2.to_f * PI/180

 v = sin((lon1 - lon2 ) /2.0 )
 w = cos(lat1 ) * cos(lat2) * v * v
 x = sin((lat1 - lat2 ) /2.0 )
 d4 = 2.0 * asin(sqrt( x*x + w ) )

 d5 = ( 180 * 60 / PI ) * d4
 return d5
end 
# print "dist"
# puts dist(56,12,52,-12)


url = "http://www.ndbc.noaa.gov/ship_obs.php?uom=M&time=#{time_span}"
#time=2 reports past two hours time=O this hour etc

wind_fact = 3600.0/1852.0 # conversion m/s to kts

open('wx_rep.gpx',"w") do |h|
  h.puts "<?xml version=\"1.0\" encoding=\"utf-8\"?>
<gpx version=\"1.1\" creator=\"Wx-ndbc.rb\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.topografix.com/GPX/1/1\" xsi:schemaLocation=\"http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd\">"

 lines = open url
 lines.each do |line|
	if line =~ /background-color.*SHIP/
	  
	  s = line.split("SHIP")[1]  
	  r = s.split(" ")
	  wind = (wind_fact * r[4].to_f).round
	  


	  if (radius != nil && latC != nil && longC != nil) && dist(latC,longC,r[1],r[2]).to_i < radius 
		
		h.puts "<wpt lat=\"#{r[1]}\" lon=\"#{r[2]}\">"
		h.puts "<name>#{r[3]},#{wind}kts,#{r[0]}UTC,#{r[8]}mb  </name>"
		h.puts "<desc></desc>"
		h.puts "<sym>xmred</sym>"
		h.puts "<type>WPT</type>"
		h.puts "<prop>A,0,1,1,1</prop>"
		h.puts " </wpt>"
		ships = ships + 1

	  elsif ( radius == nil && latC == nil && longC == nil )	
				h.puts "<wpt lat=\"#{r[1]}\" lon=\"#{r[2]}\">"
		h.puts "<name>#{r[3]},#{wind}kts,#{r[0]}UTC,#{r[8]}mb  </name>"
		h.puts "<desc></desc>"
		h.puts "<sym>xmred</sym>"
		h.puts "<type>WPT</type>"
		h.puts "<prop>A,0,1,1,1</prop>"
		h.puts " </wpt>"
		ships = ships + 1
	
	  elsif ( radius == nil || latC == nil || longC == nil ) || ARGV[4] != nil
		puts "Wrong format. Use #{$0} without argument for help."
		exit false
	  end #if radius
		
	end				#if line
  end 				#lines.each
  if (radius != nil && latC != nil && longC != nil)
  h.puts		         "<wpt lat=\"#{latC}\" lon=\"#{longC}\">
				  <name>Center</name>
				  <desc></desc>
				  <sym>triangle</sym>
				  <type>WPT</type>
				  <prop>A,0,1,1,1</prop>
		          </wpt>"
  end
  h.puts "</gpx>"
end 				#open

puts "Import 'wx_rep.gpx' into OpenCPN !"
puts "#{ships} ship reports available."

exit
