#!/usr/bin/ruby -w

longitude_degrees_between_waypoints = 5 # OK change this to suit you

############################################################################
#    Copyright (C) 2009 by Thomas Hockne                                   #
#                                                                          #
#                                                                          #
#    This program is free software; you can redistribute it and/or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################
#2009-12-12 ver 0.2
#2009-12-13 added gc distance & composite distance.
#2009-12-13 check that limiting lat < lat vertex.
#2009-12-15 cleaning up code using 'include Math' and using radians where possible.
#2009-12-16 limiting lat must be < 90 to start with.
#2009-12-27 cleaning up for ver 0.4
#2010-01-06 Adjustments to conform with OpenCPNs new standard compatible  GPX format.
#2010-01-08 Added built in version message.

include Math

GCVERSION = 0.4

if ARGV[0] =~/v|V/
  puts "GC version #{GCVERSION}"
  exit
end

unless ARGV[0] && ARGV[1] && ARGV[2] && ARGV[3]
	puts "Usage: #{$0} lat-start long-start lat-dest long-dest [limiting lat] [% long diff between wpts, default = 5]"
	puts "#{$0} -v , reports the version of the script"
 	exit(false)
end
 
TO_RADIANS = PI/180
TO_DEGREES = 180/PI
TWO_PI = 2 * PI
HALF_PI = PI/2
ONE_AND_A_HALF_PI = 1.5 * PI
 
lat_dep = ARGV[0].to_f * TO_RADIANS
long_dep = ARGV[1].to_f * TO_RADIANS
lat_dest = ARGV[2].to_f * TO_RADIANS
long_dest = ARGV[3].to_f * TO_RADIANS

if ARGV[4] =~ /%/ && ARGV[4] != nil
  longitude_degrees_between_waypoints = ARGV[4].split("%")[1].to_f
elsif ARGV[4] != nil
 lat_limit = ARGV[4].to_f  * TO_RADIANS
end

if (ARGV[5] =~ /%/ )
  longitude_degrees_between_waypoints = ARGV[5].split("%")[1].to_f
elsif ARGV[5] != nil
  puts "Ignoring long diff between wpts as it is wrong format, must be of format %3"
end

message1 = "Start"				
message2 = "Destination"
message3 = ""
error = 0
error_message = ""
course = nil
no_vertex = 0

LONG_DIFF = longitude_degrees_between_waypoints * TO_RADIANS # degrees longitude  between way-points on gc track 

# ######################## rec2polar angle #######################################

def rec2polar_angle(x,y)
  
#  r = sqrt(x*x +y*y)
  x = x.to_f
  y = y.to_f
  v = atan(y/x)
  if x > 0 && y < 0
	v = 2*PI + v
  end
  if x < 0 && y < 0
	v = PI + v
  end
  if x < 0 && y > 0
	v = PI + v
  end
     return v 
end


######################### long diff ########################################

def diff_long(longS,longD)
  longS = longS.to_f 
  longD = longD.to_f 
  
  if (longS.abs + longD.abs) <= PI
	  diff = longD -longS
	  
  elsif (longS.abs + longD.abs) > PI
	if longS > 0 && longD < 0  # heading E
	  diff = 2.0 * PI + (longD -longS)
	elsif  longS < 0 && longD > 0  # heading W
	  diff = (longD - longS) - 2.0 * PI
	elsif  ( longS > 0 && longD > 0 )||( longS < 0 && longD < 0 )
	   diff = longD -longS
	end
  end
  return diff
end

sign = 0
if diff_long(long_dep,long_dest) >= 0 #heading E
  sign = 1
elsif diff_long(long_dep,long_dest) < 0 #heading W
   sign = -1
end

################## Calculate LAT & LONG Vertexes ####################################
# c = initial course 
#Formula thanks to Bowditch Vol II

if ((long_dep * TO_DEGREES + 180).modulo(180) == (long_dest * TO_DEGREES + 180).modulo(180)) || ( lat_dest == 0 && lat_dep == 0 )
  no_vertex = 1
  error_message = error_message + " No Vertexes exists!"
end

def Course(lat_dep,long_dep,lat_dest,long_dest)
	recx = cos(lat_dep) * tan(lat_dest) - sin(lat_dep) * cos(diff_long(long_dep,long_dest))
	recy = sin(diff_long(long_dep,long_dest))
	c = rec2polar_angle(recx,recy)
    return c 
end

################ Great Circle Distance #####################

 def distance(lat_a,long_a,lat_b,long_b) #arguments in radians

   distance = acos( cos(lat_a) * cos(lat_b) *cos(diff_long(long_a,long_b)) + sin(lat_a) * sin(lat_b))

   return distance 
 end
 


unless no_vertex == 1

  course = Course(lat_dep,long_dep,lat_dest,long_dest)

  lat_vertex = acos(cos(lat_dep) * sin(course))

  if lat_vertex > HALF_PI
	lat_vertex = PI - lat_vertex      
  end
  if lat_vertex < - HALF_PI
	lat_vertex = lat_vertex + PI
  end



  
  if course > HALF_PI && course < ONE_AND_A_HALF_PI && lat_dep < 0		# Rules from Bowditch
	lat_vertex = -lat_vertex			               
  elsif course > HALF_PI && course < ONE_AND_A_HALF_PI && lat_dest < 0	
	lat_vertex = -lat_vertex
  end


  
  if lat_dep.abs <= lat_dest.abs && ((lat_dep>0 && lat_dest<0) ||(lat_dep<0 && lat_dest>0))
	
	long_vertex = long_dep + asin(cos(course)/sin(lat_vertex)) * -sign + PI
  else
	long_vertex = long_dep + asin(cos(course)/sin(lat_vertex)) * sign
  end
 


  if long_vertex > PI
	long_vertex = long_vertex - TWO_PI
  elsif long_vertex <  -PI
	long_vertex = long_vertex + TWO_PI
  end

 
end

######## The opposite Vertex ######################
unless no_vertex == 1
  
  lat_vertex2 = -lat_vertex

  if long_vertex >= 0
	long_vertex2 = long_vertex - PI
  else
	long_vertex2 = long_vertex + PI
  end
  
end
################## Evaluate arguments ##########################################

if ((long_dep * TO_DEGREES + 180).modulo(180) == (long_dest * TO_DEGREES + 180).modulo(180)) && (long_dep != long_dest)
   error = 1
   error_message = error_message + " Can't handle a Great Circle Route through the Poles."
end
 

if lat_dep.abs >= HALF_PI|| lat_dest.abs >= HALF_PI
  error = 1
  error_message = error_message + " Error Lat > 90.\n"
end
if long_dep.abs > PI || long_dest.abs > PI
  error = 1
  error_message = error_message + " Error Long > 180.\n"
end


if lat_limit.to_f.abs >= HALF_PI && lat_limit != nil
  error = 1
  error_message = error_message + " Error Limiting Lat > 90.\n"
end




if lat_limit != nil
  if sign == 1 && (diff_long(long_dest ,long_vertex)> 0 || diff_long(long_dep,long_vertex)< 0) && (diff_long(long_dest ,long_vertex + PI)> 0 || diff_long(long_dep,long_vertex + PI)< 0) && lat_limit != nil
	error =1
	error_message = error_message + " Limiting Lat don't work as vertex is not between the Start and the Destination.\n"
  end
  if sign == -1 && (diff_long(long_dest,long_vertex)< 0 || diff_long(long_dep,long_vertex)> 0) && (diff_long(long_dest,long_vertex + PI)< 0 || diff_long(long_dep,long_vertex + PI)> 0) && lat_limit != nil
	error = 1
	error_message = error_message + " Limiting Lat don't work as vertex is not between the Start and the Destination.\n"
  end
 
  

if lat_limit.to_f.abs   > lat_vertex.abs
  error = 1
  error_message = error_message + " Limiting Latitude can't be greater than the vertexes."
end


  if lat_dep < lat_dest && lat_limit.to_f  > lat_dep && lat_limit.to_f  < lat_dest
	error = 1
	error_message = error_message + " Limiting Lat must not be between the Start and the Destination\n"
	 
  elsif lat_dep > lat_dest && lat_limit.to_f < lat_dep && lat_limit.to_f > lat_dest
	error = 1
	error_message = error_message + " Limiting Lat must not be between the Start and the Destination\n"
  end
  
  if lat_dep > 0 && lat_dest>0 && lat_limit.to_f < 0
	error = 1
	error_message = error_message + " Limiting Lat has wrong sign."
  elsif lat_dep < 0 && lat_dest < 0 && lat_limit.to_f > 0
	error = 1
	error_message = error_message + " Limiting Lat has wrong sign."
  end
  
  if lat_dep < 0 && lat_dest > 0 && lat_dest.abs > lat_dep.abs && lat_limit.to_f < 0
	error = 1 
	error_message = error_message + " Limiting  Lat must be close to the numerical greater Lat."
  elsif lat_dep > 0 && lat_dest < 0 && lat_dest.abs > lat_dep.abs && lat_limit.to_f > 0
	error = 1 
	error_message = error_message + " Limiting  Lat must be close to the numerical greater Lat."
  end
  
  if lat_dep < 0 && lat_dest > 0 && lat_dest.abs < lat_dep.abs && lat_limit.to_f > 0
	error = 1 
	error_message = error_message + " Limiting  Lat must be close to the numerical greater Lat."
  elsif lat_dep > 0 && lat_dest < 0 && lat_dest.abs < lat_dep.abs && lat_limit.to_f < 0
	error = 1 
	error_message = error_message + " Limiting  Lat must be close to the numerical greater Lat."
  end
  
end

if  error == 1
  STDERR.puts error_message
  exit(3)
end

#################calculate  latitude for intermediary longitude ver 1 ##################
#Formula from http://williams.best.vwh.net/avform.htm

def inter_long1(lat_dep,long_dep,lat_dest,long_dest,long_int)
  
lat_int1 = atan((sin(lat_dep)*cos(lat_dest)*sin(long_int-long_dest)-sin(lat_dest)*cos(lat_dep)*sin(long_int-long_dep))/(cos(lat_dep)*cos(lat_dest)*sin(long_dep-long_dest)))


return lat_int1  

end


#############################Composite Sailing##############################################
#Formula thanks to Bowditch Vol II 1011 p 591 1975.


if lat_limit != nil
  
  long_diff_start = acos( tan(lat_dep)/tan(lat_limit)) # long diff start to vertex first GC
  long_diff_dest = acos( tan(lat_dest)/tan(lat_limit)) # long dest start to vertex last GC
  long_start_vertex = long_dep + long_diff_start * sign
  long_dest_vertex = long_dest - long_diff_dest * sign

end
################## Pure Great Circle Distance ##################################### 


 
 gc_distance = distance(lat_dep ,long_dep ,lat_dest ,long_dest )* TO_DEGREES * 60 #(nautical miles M)
 
################## Composite Sailing  Distance #####################################  
 
 if lat_limit != nil
   
  composite_distance = distance(lat_dep ,long_dep ,lat_limit ,long_start_vertex )* TO_DEGREES * 60 + diff_long(long_start_vertex ,long_dest_vertex ) * cos(lat_limit) * TO_DEGREES * 60 * sign + distance(lat_limit,long_dest_vertex,lat_dest ,long_dest)* TO_DEGREES * 60
  
 end
 

 
  if lat_limit == nil
    message3 = gc_distance.round
	message1 = "GC Distance"
  elsif lat_limit != nil
	message3 = composite_distance.round
	message1 = "Composite GC Distance"
  end

###################### Write GPX for Route Display #########################

open('Great_Circle.gpx',"w") do |h|
 h.puts "<?xml version=\"1.0\" encoding=\"utf-8\"?>
<gpx version=\"1.1\" creator=\"OpenCPN\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.topografix.com/GPX/1/1\" xsi:schemaLocation=\"http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd\">
        <rte>
                <name></name>
                <number>1</number>"
				
h.puts "	                <rtept lat=\"#{lat_dep* TO_DEGREES}\" lon=\"#{long_dep* TO_DEGREES}\">
                        <name>#{message1} #{message3}</name>
                        <desc></desc>
                        <sym>diamond</sym>
                        <type>WPT</type>
                        <prop>A,1,1,0,1</prop>
                </rtept>"	
				
if 	lat_limit == nil && error == 0		
######################################### Pure GC Sailing ##################################				
long_int = long_dep
name = 0

while diff_long(long_int,long_dest ).abs  > LONG_DIFF 
  
  lat_int = inter_long1(lat_dep,long_dep,lat_dest,long_dest,long_int)
		
  
            h.puts  "<rtept lat=\"#{lat_int * TO_DEGREES}\" lon=\"#{long_int * TO_DEGREES}\">
                        <name>#{name}</name>
                        <desc></desc>
                        <sym>xmred</sym>
                        <type>WPT</type>
                        <prop>A,1,1,0,1</prop>
                </rtept>"				
		name = name + 1	
		long_int = 	long_int + LONG_DIFF * sign

end				
####################################### END PURE GC #####################################
elsif lat_limit != nil && error == 0
#========================================Composite Sailing==============================

# lat_limit = lat_limit
 long_int = long_dep
 name = 0

while diff_long(long_int ,long_start_vertex ).abs  > LONG_DIFF 
  
  lat_int = inter_long1(lat_dep,long_dep,lat_limit,long_start_vertex,long_int)
		
  
            h.puts  "<rtept lat=\"#{lat_int * TO_DEGREES}\" lon=\"#{long_int * TO_DEGREES}\">
                        <name>#{name}</name>
                        <desc></desc>
                        <sym>xmred</sym>
                        <type>WPT</type>
                        <prop>A,1,1,0,1</prop>
                </rtept>"				
		name = name + 1	
		long_int = 	long_int + LONG_DIFF * sign

end	 
h.puts "                <rtept lat=\"#{lat_limit * TO_DEGREES}\" lon=\"#{long_start_vertex* TO_DEGREES}\">
                        <name>#{name}</name>
                        <desc></desc>
                        <sym>xmred</sym>
                        <type>WPT</type>
                        <prop>A,1,1,0,1</prop>
                        </rtept> "  

long_int = long_dest_vertex
name  = name + 1	

while diff_long(long_int ,long_dest).abs > LONG_DIFF 
  
  lat_int = inter_long1(lat_limit,long_dest_vertex,lat_dest,long_dest,long_int)
		
  
            h.puts  "<rtept lat=\"#{lat_int * TO_DEGREES}\" lon=\"#{long_int * TO_DEGREES}\">
                        <name>#{name}</name>
                        <desc></desc>
                        <sym>xmred</sym>
                        <type>WPT</type>
                        <prop>A,1,1,0,1</prop>
                </rtept>"

		name = name + 1	
		long_int = 	long_int + LONG_DIFF * sign

end							
 elsif  error == 0
  puts "Can't happen!?"
  puts lat_limit 
end  
######################################### END Composite Sailing ######################### 


  
h.puts "                <rtept lat=\"#{lat_dest* TO_DEGREES}\" lon=\"#{long_dest* TO_DEGREES}\">
                        <name>#{message2}</name>
                        <desc></desc>
                        <sym>diamond</sym>
                        <type>WPT</type>
                        <prop>A,1,1,0,1</prop>
                        </rtept> "
 
 
h.puts "        </rte>
</gpx>" 
end

################################## Show Vertexes as independent WPT:s ####################

if  error == 0 && no_vertex == 0
  open('Great_Circle_vertex.gpx',"w") do |g|
  g.puts       "<?xml version=\"1.0\" encoding=\"utf-8\"?>
  <gpx version=\"1.1\" creator=\"OpenCPN\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns=\"http://www.topografix.com/GPX/1/1\" xsi:schemaLocation=\"http://www.topografix.com/GPX/1/1 http://www.topografix.com/GPX/1/1/gpx.xsd\">
		
		         <wpt lat=\"#{lat_vertex * TO_DEGREES}\" lon=\"#{long_vertex* TO_DEGREES}\">
				  <name>VERTEX_1</name>
				  <desc></desc>
				  <sym>triangle</sym>
				  <type>WPT</type>
				  <prop>A,0,1,1,1</prop>
		          </wpt>
				  
				  <wpt lat=\"#{lat_vertex2 * TO_DEGREES}\" lon=\"#{long_vertex2 * TO_DEGREES}\">
				  <name>VERTEX_2</name>
				  <desc></desc>
				  <sym>triangle</sym>
				  <type>WPT</type>
				  <prop>A,0,1,1,1</prop>
		          </wpt>
				 
		  
  </gpx>"
  
  
  end #if
  
end #open
  
if no_vertex == 1
  File.unlink 'Great_Circle_vertex.gpx' if File.exists?('Great_Circle_vertex.gpx')
end

############################### Out-put to CLI ###########################################

puts " Great Circle route data in Great_Circle.gpx"
puts " Vertex wpts in Great_Circle_vertex.gpx " unless no_vertex == 1
puts error_message if no_vertex == 1
puts " Use OpenCPN GPX-import to display."
puts " Using Long diff = #{longitude_degrees_between_waypoints} in gpx file."
puts " Calculated Great Circle Distance is #{gc_distance.round} M." if lat_limit == nil
puts " Composite Sailing Distance is #{composite_distance.round} M." unless lat_limit == nil


exit